def inpsect_none_if_exception_or_empty(obj_to_inspect, inspect_fct_name):
    """runs an inspect function on obj_to_inspect and returns None if the result is empty or returns an exception"""
    import inspect

    inspect_fct = getattr(inspect, inspect_fct_name)
    ret = None
    try:
        ret = inspect_fct(obj_to_inspect)
        if len(ret) == 0: ret = None
    except Exception as e:
        pass

    return ret

def gettime():
    """return timestamp"""
    import time
    return time.time()

def get_hash_key(to_hash:str, prefix=None, suffix=None):
    """return a hash of to_hash that can serve as a key"""

    import hashlib
    if prefix is None:
        prefix = ""

    if suffix is None:
        suffix = ""

    val =  prefix + str(hashlib.sha256(to_hash.encode("utf-8")).hexdigest()) + suffix
    return legalize_key( val )

def legalize_key(key):
    """returns a string that can serve as a valid key for database"""
    import re
    key = key.lower().replace(" ", "_")

    pattern = re.compile('[\W]+') 
    key = pattern.sub('', key) 

    return key

def getuid():
    """returns a random id that can serve as a key"""
    import uuid
    val = str(uuid.uuid4())
    return legalize_key(val)

def get_random_variable_name():
    """returns a string that can serve as a python variable name"""
    import uuid
    val = str(uuid.uuid4())
    return "var_" + legalize_key(val)


from pyArango.collection import Collection, Edges, Field
import pyArango.validation as VAL
import pyArango.graph as GR


COLLECTIONS = ["Jobs", "MachineElves", "MachineElvesRevisions", "Failures", "JobFailures", "Parameters", "JobParameters", "Results", "JobToJob"]
GRAPHS = ["Jobs_graph", "JobFailures_graph", "JobParameters_graph"]

class Jobs(Collection) :
    """Schema of a job in the database"""
    _fields = {
        "task" : {
            "name": Field(validators = [VAL.NotNull()]),
            "signature": Field(validators = [VAL.NotNull()]),
            "source_code": Field(),
            "documentation": Field(),
            "revision": Field(),
        },
        "machine_elf" : {
            "id": Field(validators = [VAL.NotNull()]),
            "documentation": Field(),
            "revision": Field(),
        },
        "parameter_ids": Field(default=list),
        "submit_date" : Field(validators = [VAL.NotNull()]),
        "start_date": Field(),
        "end_date": Field(),
        "status": Field(validators = [VAL.NotNull()]),
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class Failures(Collection) :
    """Schema of a failure in the database"""

    _fields = {
        "type": Field(validators = [VAL.NotNull()]),
        "value": Field(validators = [VAL.NotNull()]),
        "traceback": Field(validators = [VAL.NotNull()]),
        "creation_date": Field(validators = [VAL.NotNull()])
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class MachineElves(Collection) :
    """Schema of a machine elf in the database"""

    _fields = {
        "documentation": Field(),
        "last_revision": Field(),
        "creation_date" : Field(validators = [VAL.NotNull()]),
        "revisions": {
            "dates": Field(default=list),
            "hashes": Field(default=list),
        }
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class MachineElvesRevisions(Collection) :
    """Schema of a machine elf revision in the database"""

    _fields = {
        "documentation": Field(),
        "creation_date" : Field(validators = [VAL.NotNull()]),
        "source_code": Field()
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class Parameters(Collection) :
    """Schema of a parameter in the database. A parameter is a variable passed as an argument to a task"""

    _fields = {
        "creation_date": Field(validators=[VAL.NotNull()]),
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": True
    }

class Results(Collection) :
    """Schema of a result in the database"""

    _fields = {
        "creation_date": Field(validators=[VAL.NotNull()]),
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": True
    }

class JobParameters(Edges) :
    """Schema of a job parameter in the database. a job parameter is an edge that associates a parameter to task argument"""

    _fields = {
        "creation_date": Field(validators=[VAL.NotNull()]),
        "name": Field(),
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class JobToJob(Edges) :
    """Edge connecting jobs that depend on each other"""
    _fields = {
        "creation_date": Field(validators=[VAL.NotNull()]),
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class JobParameters_graph(GR.Graph):
    """Graph connecting jobs to parameters"""
    _edgeDefinitions = (
        GR.EdgeDefinition("JobParameters", fromCollections = ["Jobs"], toCollections = ["Parameters"]),
    ) 
    _orphanedCollections = []

class JobFailures(Edges) :
    """Edge connecting jobs to failures"""
    _fields = {
        "creation_date" : Field(validators = [VAL.NotNull()])
    }

    _validation = {
        "on_save": True,
        "on_set": False,
        "allow_foreign_fields": False
    }

class Jobs_graph(GR.Graph):
    """The job orchestration graph"""
    _edgeDefinitions = (
        GR.EdgeDefinition("JobToJob", fromCollections = ["Jobs"], toCollections = ["Jobs"]),
    ) 
    _orphanedCollections = []

class JobFailures_graph(GR.Graph):
    """Graph connecting jobs to failures"""
    _edgeDefinitions = (
        GR.EdgeDefinition("JobFailures", fromCollections = ["Jobs"], toCollections = ["Failures"]),
    ) 
    _orphanedCollections = []




from . import utils as ut
from . import custom_types
from . import models as mod
from .the_exceptions import *


import pyArango.theExceptions as a_exc

from icecream import ic
ic.configureOutput(includeContext=True)

import logging
logger = logging.getLogger("BLUMYCELIUM")

class ArangoMycelium:
    """docstring for ArangoMycelium. A mycelium over ArangoDB"""

    def __init__(self, connection, name):
        """
        connection: pyArango connection object.
        name: Unique name for the mycelium
        """

        self.connection = connection
        self.name = name
        self.db_name = "BLUMYC_" + name

        self.collections = mod.COLLECTIONS
        self.graphs = mod.GRAPHS

        self.db = None

        if not self.connection.hasDatabase(self.db_name):
            logger.warning("Warning: Database %s does not exist. To create it run self.init with init_db=True" % self.db_name)
        else:
            self.db = self.connection[self.db_name]

    def _init_db(self) :
        """
        Create the mycelium database
        """
        from pyArango.theExceptions import CreationError

        print("Creating database", self.db_name)
        try :
            return self.connection.createDatabase(self.db_name)
        except CreationError as e :
            print("Database Creation error %s: => %s" % (self.db_name, e.message))

            try :
                return self.connection[self.db_name]
            except :
                raise e

    def init(self, init_db=False, users_to_create=None) :
        """
        init_db: initialise the database
        users_to_create: list of dicts {username, password}
        """
        if init_db:
            self.db = self._init_db()

        if self.db is None:
            raise DatabaseNotFoundError("Cannot continue with the initialisation because database does not exist. Try runing init with init_db=True")

        logger.info("Initializing %s..." % self.db_name)    
        logger.info("-- init collections")
        self._init_collections(self.db)
        logger.info("-- init graphs")
        self._init_graphs(self.db)
        logger.info("-- init indexes")
        self._init_indexes(self.db)
        logger.info("-- init users")

        if users_to_create:
            for user in self.users_to_create :
                self._init_user(self.connection, self.db_name, user["username"], user["password"])
    
    def drop_jobs(self):
        """delete all information related to jobs"""
        for collection in ["Jobs", "Failures", "Parameters", "JobFailures", "JobParameters", "Results", "JobToJob"]:
            self.db[collection].truncate() 

    def drop(self):
        """delete all documents in the mycelium"""
        for collection in self.collections:
            self.db[collection].truncate() 

    def _init_collections(self, db) :
        """initialize all collections"""
        for collection in self.collections :
            if collection.lower() not in ("collection", "edges", "field") :
                logger.info("Creating collection", collection)
                try :
                    db.createCollection(collection, allowUserKeys=True)
                except Exception as exp :
                    logger.info("Collection Creation error: =>", exp)

    def _init_graphs(self, db) :
        """initialize all graphs"""

        for graph in self.graphs :
            # ic(graph)
            if graph.lower() != "graph":
                logger.info("Creating graph", graph)
                try :
                    db.createGraph(graph)
                except Exception as exp :
                    logger.info("Graph Creation error (%s): =>" %graph, exp)
    
    def _init_user(self, connection, dbName, username, password) :
        """initialize all users"""

        logger.info("creating user: %s" % username)
        
        user = connection.users.createUser(username, password)
        user["extra"] = "Created on: %s, by automatic setup" % ut.gettime()
        try :
            user.save()
        except Exception as e1 :
            logger.info("Can't create user: '%s' -> " % username, e1)
            try :
                user = connection.users.fetchUser(username)
            except Exception as e2 :
                raise UserCreationError("Unable to get user: '%s' -> " % username, e2)
        u.setPermissions(dbName, True)

    def _init_indexes(self, *args, **kwargs):
        """initialize all indexes"""
        logger.info("No indexes defined")

    def register_machine_elf(self, machine_elf, store_source):
        """register and elf in the mycellium"""
        now = ut.gettime()
        first_register = False
        try:
            elf_doc = self.db["MachineElves"][machine_elf.uid]
            elf_doc["revisions"]["dates"].append(now)
            elf_doc["revisions"]["dates"] = elf_doc["revisions"]["dates"]
            elf_doc["revisions"]["hashes"].append(machine_elf.revision)
            elf_doc["revisions"]["hashes"] = elf_doc["revisions"]["hashes"]
        except a_exc.DocumentNotFoundError:
            elf_doc = self.db["MachineElves"].createDocument()
            elf_doc["_key"] = machine_elf.uid
            elf_doc["creation_date"] = now
            elf_doc["revisions"]["dates"] = [elf_doc["creation_date"]]
            elf_doc["revisions"]["hashes"] = [machine_elf.revision]
            first_register = True

        revision_key = ut.legalize_key(machine_elf.revision)

        if store_source and (elf_doc["last_revision"] != machine_elf.revision or first_register) and (self.db["MachineElvesRevisions"]):
            revision_doc = self.db["MachineElvesRevisions"].createDocument()
            revision_doc["_key"] = revision_key
            revision_doc["source_code"] = machine_elf.source_code
            revision_doc["creation_date"] = now
            revision_doc.save()

        elf_doc["documentation"] = machine_elf.documentation
        elf_doc["last_revision"] = machine_elf.revision
        
        elf_doc.save()

    def get_job(self, job_id):
        """get a job"""
        return self.db["Jobs"][job_id]
    
    def _save_job(self, job, now_date):
        """save job to the mycelium"""
        job_key = ut.legalize_key(job.run_job_id)
        job_doc = self.db["Jobs"].createDocument()

        job_doc.set(
            {
            "_key": job_key,
            "task" : {
                    "name": job.task.name,
                    "signature": str(job.task.signature),
                    "source_code": job.task.source_code,
                    "documentation": job.task.documentation,
                    "revision": job.task.revision,
                },
                "machine_elf" : {
                    "id": job.worker_elf.uid,
                    "revision": job.worker_elf.revision,
                },
                # "static_parameters": job.parameters.get_static_parameters(),
                "submit_date" : now_date,
                "start_date": None,
                "end_date": None,
                "status": custom_types.STATUS["PENDING"],
            }
        )
        job_doc.save()
        return job_doc

    def _save_parameters(self, job_doc, params, now):
        """save a parameters"""
        # from rich import print
        def _get_param_doc(node, date):
            doc = self.db["Parameters"].createDocument()
            dct = {"creation_date": date}
            dct.update(node)
            doc.set(dct)
            return doc

        graph = self.db.graphs["JobParameters_graph"]
        for name, traversal in params.items():
            param_doc = _get_param_doc(traversal, now)
            data = {"creation_date": now, "name": name}
            graph.link("JobParameters", job_doc, param_doc, data)

    def push_job(self, job):
        """push a job to the mycelium"""
        # from rich import print
        now = ut.gettime()
        job_doc = self._save_job(job, now)
        
        params = job.parameters.get_parameter_dict()
        self._save_parameters(job_doc, params, now)
        
        graph = self.db.graphs["Jobs_graph"]
        if job.dependencies:
            for dep in job.dependencies:
                graph.link("JobToJob", "Jobs/" + dep, job_doc, {"creation_date": now})

    def get_job_parameters(self, job_id):
        """return the parameters for a job"""
        aql = """
        FOR jparam IN JobParameters
            FILTER jparam._from == @id
            FOR node IN Parameters
                FILTER node._id == jparam._to
                RETURN {
                    name: jparam.name,
                    node: node
                }
        """
        bind_vars = {"id": "Jobs/"+ job_id}
        res_aql = self.db.AQLQuery(aql, bindVars=bind_vars, batchSize=100, rawResults=True)
        ret = {}
        for res in res_aql:
            ret[res["name"]] = res["node"]

        return ret

    def get_jobs(self, elf_uid:str, all_jobs=False, status_restriction=[custom_types.STATUS["PENDING"]]):
        """return all jobs for an elf"""
        bind_vars = {"uid": elf_uid}
        if all_jobs:
            status_restriction = list(custom_types.STATUS.values())
        
        str_status_filter = []
        for status in status_restriction:
            pending = "job.status == @%s" % status
            str_status_filter.append(pending)
            bind_vars[status] = status

        if len(str_status_filter) > 0:
            str_status_filter = "FILTER " + " OR ".join(str_status_filter)
        else:
            str_status_filter = ""

        aql = """
            FOR job IN Jobs
                FILTER job.machine_elf.id == @uid
                {str_status_filter}
                SORT job.creation_date DESC
                RETURN job
        """.format(str_status_filter=str_status_filter)

        ret_q = self.db.AQLQuery(aql, bindVars=bind_vars, batchSize=100, rawResults=True)
        ret = []
        for job in ret_q:
            job["id"] = job["_key"]
            ret.append(job)
        return ret

    def is_job_ready(self, job_id):
        """return True if the job is ready to run"""
        job_doc = self.get_job(job_id)
        if job_doc["status"] not in [ custom_types.STATUS["PENDING"], custom_types.STATUS["READY"] ]:
            return False

        aql = """
            FOR job IN Jobs
                FILTER job._id == @id
                FOR jtj in JobToJob
                    FILTER jtj._to == job._id
                    FOR job2 IN Jobs
                        FILTER job2._id == jtj._from
                        RETURN job2
        """

        ready = 0
        count = 0
        ret_q = self.db.AQLQuery(aql, bindVars={"id": job_doc["_id"]}, batchSize=100, rawResults=True)
        for job in ret_q:
            count += 1
            if job["status"] == custom_types.STATUS["DONE"]:
                ready += 1

        self_ready = count == ready
        return self_ready

    def update_job_status(self, job_id, status):
        """update the status of a job"""
        job_doc = self.get_job(job_id)
        job_doc["status"] = status
        job_doc.save()

    def start_job(self, job_id):
        """mark a job as running"""
        job_doc = self.get_job(job_id)
        job_doc["status"] = custom_types.STATUS["RUNING"]
        job_doc["start_date"] = ut.gettime()
        job_doc.save()

    def complete_job(self, job_id):
        """mark a job as successful"""
        job_doc = self.get_job(job_id)
        job_doc["status"] = custom_types.STATUS["DONE"]
        job_doc["end_date"] = ut.gettime()
        job_doc.save()

    def register_job_failure(self, exc_type, exc_value, exc_traceback, job_id):
        """register a job failure"""
        import traceback
        import hashlib

        e = traceback.extract_tb(exc_traceback)
        now = ut.gettime()

        self.update_job_status(job_id, custom_types.STATUS["FAILED"])
        
        trace = traceback.extract_tb(exc_traceback).format()
        trace_str = "".join(trace).encode("utf-8")
        trace_key = ut.legalize_key( str( hashlib.sha256(trace_str).hexdigest() ) )

        try:
            failure_doc = self.db["Failures"][trace_key]
        except a_exc.DocumentNotFoundError:
            failure_doc = self.db["Failures"].createDocument()
            failure_doc.set(
                {
                    "_key": trace_key,
                    "type": str(exc_type),
                    "value": str(exc_value),
                    "traceback": trace,
                    "creation_date": now
                }
            )
            failure_doc.save()

        job_doc = self.get_job(job_id)
        graph = self.db.graphs["JobFailures_graph"]
        graph.link("JobFailures", job_doc, failure_doc, {"creation_date": now})
        job_doc["status"] = custom_types.STATUS["FAILED"]
        job_doc["end_date"] = ut.gettime()
        job_doc.save()

    def store_results(self, job_id, results:dict):
        """store the results of a job"""
        if results is None:
            return
        
        if not type(results) is dict:
            raise  ResultTypeError("Results must be None or a dictionary")
        
        now = ut.gettime()

        for name, value in results.items():
            result_key = value["result_id"]
            try:
                result_doc = self.db["Results"][result_key]
            except a_exc.DocumentNotFoundError:
                result_doc = self.db["Results"].createDocument()
        
            result_doc.set(
                {
                    "_key": value["result_id"],
                    "name": name,
                    "value": value["value"],
                    "creation_date": now,
                }
            )
            result_doc.save()

    def get_result(self, result_id):
        """return the result of a job"""
        try:
            result_doc = self.db["Results"][result_id]
        except a_exc.DocumentNotFoundError:
            raise ResultNotFound("Unable to retrieve result: %s" % result_id)

        return result_doc.getStore()["value"]

    def get_job_status(self, job_id):
        """return the status of a job"""
        job_doc = self.db["Jobs"][job_id]
        return job_doc.getStore()["status"]

    def get_dependencies_status(self, job_id):
        aql = """
            FOR jj IN JobToJob
                FILTER jj._to == @id
                FOR job IN Jobs
                    FILTER jj._from == job._id
                    RETURN {
                        "job_id": job._key,
                        "status": job.status
                    }
        """
        
        bind_vars = {"id": "Jobs/"+ job_id}
        res_q = self.db.AQLQuery(aql, bindVars=bind_vars, batchSize=100, rawResults=True)
        ret = [res for res in res_q]
        return ret




from . import utils as ut
# import utils as ut
from icecream import ic
ic.configureOutput(includeContext=True)

class ExecutionError(Exception):
    def __init__(self, code):
        msg="===\n{code}\n===".format(code=code)
        self.message = msg

class CodeBlock:
    """a code block of python code to run"""
    def __init__(self, init_code=None, return_statement=None):
        self.init_code = init_code
        self.return_statement = return_statement

    def format(self, **string_kwargs):
        """replace the name of variables by their myc_back_variables references"""
        import re
        def _sub_var_names(code):
            ret = re.sub("(var_[^\W]+)", r"myc_back_variables['\1']", code)
            return ret

        if not self.init_code is None:
            self.init_code = self.init_code.format(**string_kwargs)
            self.init_code = _sub_var_names(self.init_code)
        if not self.return_statement is None:
            self.return_statement = self.return_statement.format(**string_kwargs)
            self.return_statement =_sub_var_names(self.return_statement)

    def run(self, **myc_back_variables):
        """run the code block"""
        if not self.init_code is None:
            try:
                exec(self.init_code)
            except:
                raise ExecutionError(self.init_code)

        if not self.return_statement is None:
            try:
                return eval(self.return_statement) 
            except:
                raise ExecutionError(self.return_statement)
    
    def to_dict(self):
        """return a dictionary version of the code block"""
        return {
            "init_code": self.init_code,
            "return_statement": self.return_statement
        }

    def __str__(self):
        return "*-CodeBlock %s-*" % (id(self))

    def __repr__(self):
        str_val="""%s\n%s""" % (self.init_code, self.return_statement)
        return str_val

class GraphParameter:
    """A graph of Parameters"""
    def __init__(self, uid=None):
        if uid is None:
            uid = ut.get_random_variable_name()
        self.python_id = id(self)
        self.uid = uid
        self.value = None
        self.code_block = None
        self.dependencies = None
        self.dependency_values = {}
        
        self.computed_value = None
        
        self.origin = None
        self.pull_origin_function = None

    def set_origin(self, uid, pull_origin_function=None):
        """set the origin (address) of the parameter and function to retreive the value from the origin"""
        self.origin = uid
        self.pull_origin_function = pull_origin_function
        # ic(self, self.origin, self.pull_origin_function)

    def set_pull_origin_function(self, fct):
        """set the function to pull jvalue form the origin"""
        self.pull_origin_function = fct

    def _dict_representation(self):
        """return a dcitionnary representation of the paramater"""
        ret = {
            "uid": self.uid,
            "value": self.value,
            "code_block": self.code_block,
            "origin": self.origin
        }
        
        return ret

    def to_dict(self, reccursive=False, visited_nodes=None, copy_values=True):
        """return a dict representation of the parameter recursively if asked to"""
        import copy

        if visited_nodes is None:
            visited_nodes = set()

        value = self.value
        if copy_values:
            value = copy.copy(value)

        ret = self._dict_representation()
        
        if not self.code_block is None:
            ret["code_block"] = self.code_block.to_dict()

        if reccursive and not (self.uid in visited_nodes):
            visited_nodes.add(self.uid)
            ret["dependencies"] = {}
            if self.dependencies is not None:
                for batch in self.dependencies:
                    for uid, dep in batch.items():
                        ret["dependencies"][uid] = dep.to_dict(reccursive=True, visited_nodes=visited_nodes)
        return ret

    def add_dependencies(self, *deps):
        """add dependencies (other GraphParaemters) needed to compute the value"""
        if self.dependencies is None:
            self.dependencies = []

        if len(deps) > 0:
            dct = {dep.uid: dep for dep in deps}
            self.dependencies.append(dct)
            for duid in dct:
                self.dependency_values[duid] = None

    def make(self, visited_nodes=None, is_root=False):
        """compute the value of the paraneter and return it"""
        def _run_deps(visited):
            if self.dependencies:
                for batch in self.dependencies:
                    for dep in batch.values():
                        dep_ret = dep.make(visited)
                        if dep.uid in self.dependency_values:
                            self.dependency_values[dep.uid] = dep_ret
        
        if not self.value is None:
            self.computed_value = self.value
               
        if visited_nodes is None:
            visited_nodes = set()
        
        if self.uid in visited_nodes:
            return self.computed_value

        visited_nodes.add(self.uid)
        _run_deps(visited_nodes)

        if self.value is None:
            if not self.code_block is None:
                self.computed_value = self.code_block.run(**self.dependency_values)
            elif (not self.pull_origin_function is None ) and (not self.origin is None):
                self.computed_value = self.pull_origin_function(self.origin)
            else:
                raise ValueDerivationError("Unable to retrieve value. No defined value, code_block or origin function")
        return self.computed_value

    def traverse(self, visited_nodes=None, is_root=True, root_uid=None, to_dict=True):
        """traverse the graph dependency tree and return a dictionary representing it"""
        def _add_tree(node, is_root, root_uid, visited, as_dict):
            nnn = node
            if as_dict:
                nnn = node.to_dict()
            return {"node": nnn, "uid": node.uid, "is_root": is_root, "root_uid": root_uid, "visited": visited, "dependencies": {}}

        def _run_deps(visited, tree, tree_root_uid, as_dict):
            if self.dependencies:
                for batch in self.dependencies:
                    for dep in batch.values():
                        tree["dependencies"][dep.uid] = dep.traverse(visited_nodes=visited, is_root=False, root_uid=tree_root_uid, to_dict=as_dict)
        
        if visited_nodes is None:
            visited_nodes = set()

        if is_root:
            root_uid = self.uid

        tree = _add_tree(self, root_uid==self.uid, root_uid, self.uid in visited_nodes, as_dict=to_dict)
        
        if self.uid in visited_nodes:
            return tree

        visited_nodes.add(self.uid)
        _run_deps(visited_nodes, tree, root_uid, as_dict=to_dict)

        return tree

    @classmethod
    def build_from_traversal(cls, trav:dict, pull_origin_function=None):
        """build a parameter from a traversal dictionary"""
        def _build_node(dct_node):
            node = GraphParameter(uid=dct_node["uid"])
            node.value = dct_node["value"]
            if not dct_node["code_block"] is None:
                code_block = CodeBlock(init_code=dct_node["code_block"]["init_code"], return_statement=dct_node["code_block"]["return_statement"])
                node.code_block = code_block
            if not dct_node["origin"] is None:
                node.set_origin(dct_node["origin"])
                node.set_pull_origin_function(pull_origin_function)
            return node

        def _get_node(dct, all_nodes_dct):
            try:
                return all_nodes_dct[dct["uid"]]
            except KeyError:
                node = _build_node(dct)
                all_nodes_dct[node.uid] = node
                return node

        def _rec_build(dct, all_nodes_dct):
            if all_nodes_dct is None:
                all_nodes_dct = {}

            root_node = _get_node(dct["node"], all_nodes_dct)
            deps = []
            for dep_dct in dct["dependencies"].values():
                dep_node = _get_node(dep_dct["node"], all_nodes_dct)
                deps.append(dep_node)
                if not dct["visited"]:
                    _rec_build(dep_dct, all_nodes_dct)
            root_node.add_dependencies(*deps)
            return all_nodes_dct
        
        nodes = _rec_build(trav, None)
        root = nodes[trav["uid"]]
    
        return root

    def pp_traverse(self, full_representation=False, representation_attributes=["value", "code_block"], print_it=True):
        """a pretty print of the graph representation with dependencies"""
        from rich.tree import Tree
        from rich import print

        def _get_represenation(node):
            if full_representation:
                return str(node)
            elif representation_attributes and len(representation_attributes) > 0:
                vals = [ "%s: %s" % (attr, str( getattr(node, attr)) ) for attr in representation_attributes if not getattr(node, attr) is None]
                return ", ".join(vals)
            return repr(node)
            
        def _get_node(node):
            if node["is_root"]:
                tree = Tree("[bold blue]>ROOT: %s" % _get_represenation(node["node"]))
            elif node['visited']:
                tree = Tree("[bold magenta]>ALREADY VISITED: %s" % _get_represenation(node["node"]))
            else:
                tree = Tree("%s" % _get_represenation(node["node"]))
            return tree
        
        def _traverse(trav, curr_tree):
            for branch in trav["dependencies"].values():
                node = _get_node(branch)
                curr_tree.add(node)
                _traverse(branch, node)
            return tree

        trav = self.traverse(to_dict=False)
        tree = _get_node(trav)
        tree = _traverse(trav, tree)

        if print_it:
            print(tree)
    
        return tree

    def set_value(self, value):
        """set the static value of the parameter"""
        if not (self.code_block is None and self.origin is None) :
            raise MultipleDerivationError("Can either have a code block, a value or an origin")
        self.value = value

    def set_code_block(self, init_code, return_statement, **dependencies):
        """set the code block of the parameter"""
        if not (self.value is None and self.origin is None) :
            raise MultipleDerivationError("Can either have a code block, a value or an origin")

        string_kwargs = {}
        for hr_name, dep in dependencies.items():
            self.dependency_values[dep.uid] = None
            if not isinstance(dep, GraphParameter):
                raise DependencyTypeError("All depenedencies must be GraphParameters, got: '%s'" % type(dep))
            string_kwargs[hr_name] = dep.uid 

        if self.dependencies is None:
            self.dependencies = []
        self.dependencies.append(dependencies)
        
        
        self.code_block = CodeBlock(init_code, return_statement)
        self.code_block.format(**string_kwargs)

    def __str__(self):
        if self.dependencies is not None:
            deps = []
            for batch in self.dependencies:
                for dep in batch.values():
                    deps.append(dep.uid)
        else:
            deps = None
        return "*-GraphParameter '%s' value:'%s' code_block:'%s' origin:'%s' dependencies:'%s' -*" % (self.uid, self.value, self.code_block, self.origin, deps)

    def __repr__(self):
        deps = 0
        if self.dependencies is not None:
            for batch in self.dependencies:
                for dep in batch.values():
                    deps += 1 
        return "*-GraphParameter '%s' value:'%s' code_block:'%s' origin:'%s' dependencies:'%s' -*" % (self.uid, self.value, self.code_block, self.origin, deps)

class Value(object):
    """A wrapper for a graph parameter"""
    def _init(self, as_type=None, parent=None):
        super(Value, self).__init__()
        self.parameter = GraphParameter()
        self.as_type = as_type
        self.parent = parent
        # self.closed_init = False
        self.dependencies = []
        self.made_value = None

    # def close_init(self):
        # self.closed_init = True

    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        # self.close_init()

    def set_value(self, value):
        """set the static value"""
        self.as_type = type(value)
        return self.parameter.set_value(value)

    def set_code_block(self, *args, **kwargs):
        """set the code block to run"""
        return self.parameter.set_code_block(*args, **kwargs)

    def traverse(self, *args, **kwargs):
        """return a dictionary representing the value and it's dependencies"""
        return self.parameter.traverse(*args, **kwargs)

    def to_dict(self, *args, **kwargs):
        """return a dictionary representing the value"""
        return self.parameter.to_dict( *args, **kwargs)

    def pp_traverse(self, *args, **kwargs):
        """pretty print the value and it's dependencies"""
        return self.parameter.pp_traverse(*args, **kwargs)

    def _get_parameter(self, param):
        """return a GraphParameter"""
        if isinstance(param, Value):
            # self.parent.dependencies.append(param)
            self.parent.parameter.add_dependencies(param.parameter)
            return param.parameter
        elif isinstance(param, GraphParameter):
            self.parent.dependencies.append(param)
            self.parent.parameter.add_dependencies(param)
            return param

        new_param = GraphParameter()
        if type(param) is str:
            param = param
        new_param.set_value(param)
        return new_param

    def _validate_type(self, key):
        """is as_type is defined validate that the object specified in 'as_type' has an attribute named 'key'"""
        as_type = object.__getattribute__(self, "as_type")
        if not as_type is None:
            getattr(as_type, key) #will raise an exception if the attribute is non-existant

    def __getitem__(self, key):
        self._validate_type("__getitem__")

        code = "{self_param}[{key}]"
        key_param = GraphParameter()
        key_param.set_value(key)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, key=key_param)
        return new_param

    def __setitem__(self, key, value):
        self._validate_type("__setitem__")

        init_code="""
        def add(dct, key, value):
            dct[key]=value
            return dct""".strip()
        return_statement = "add({self_param}, {key}, {value})"
       
        key_param = self._get_parameter(key)
        value_param = self._get_parameter(value)
        
        new_param = GraphParameter()
        new_param.set_code_block(init_code=init_code, return_statement=return_statement, self_param=self.parameter, key=key_param, value=value_param)
        self.parameter = new_param

    # def __setattr__(self, key, value):
    #     # ic(key, value)
    #     try:
    #         closed_init = object.__getattribute__(self, "closed_init")
    #     except AttributeError:
    #         closed_init = False

    #     if not closed_init:
    #         return object.__setattr__(self, key, value)
    #     else:        
    #         init_code="""
    #         def add(obj, key, value):
    #             setattr(obj, key, value)
    #             return obj""".strip()
    #         return_statement = "add({self_param}, {key}, {value})"

    #         code = "{self_param}[{key}]={value}"
            
    #         key_param = self._get_parameter(key)
    #         value_param = self._get_parameter(value)
            
    #         new_param = GraphParameter()
    #         new_param.set_code_block(init_code=init_code, return_statement=return_statement, self_param=self.parameter, key=key_param, value=value_param)
            
    #         self.parameter = new_param

    def __getattr__(self, key):
        _validate_type = object.__getattribute__(self, "_validate_type")
        _validate_type(key)

        self_param = object.__getattribute__(self, "parameter")
        key_param = object.__getattribute__(self, "_get_parameter")(key)

        return_statement = "getattr({self_param}, {key})"

        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=return_statement, self_param=self_param, key=key_param)
        
        new_value = Value(parent=self)
        new_value.parameter = new_param
        return new_value

    def __call__(self, *args, **kwargs):
        def _populate(iterator, f_params, f_map, with_equals):
            for key, value in iterator:
                name = "{val_%s}" % key 
                
                if with_equals:
                    final_params.append( "%s=" % key + name)
                else:
                    final_params.append(name)
                
                final_params_map[name[1:-1]] = self._get_parameter(value)
            
            return f_params, f_map

        
        final_params = []
        final_params_map = {}
        final_params, final_params_map = _populate(enumerate(args), final_params, final_params_map, False)
        final_params, final_params_map = _populate(kwargs.items(), final_params, final_params_map, True)

        final_params = ", ".join(final_params)

        return_statement = "{self_param}(%s)" % (final_params)
        
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=return_statement, self_param=self.parameter, **final_params_map)
        
        if not self.parent is None:
            self.parent.parameter.add_dependencies(new_param)
        self.parameter = new_param

        return self

    def make(self, force=False):
        """compute a return the value represented by the Value object"""
        if self.made_value is None or force :
            self.made_value = self.parameter.make(is_root=True)
        return self.made_value

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, str(self.parameter))

    def __repr__(self):
        return "%s: %s" % (self.__class__.__name__, repr(self.parameter))

    def __mul__(self, value):
        self._validate_type("__mul__")

        code = "{self_param} * {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

    def __add__(self, value):
        self._validate_type("__add__")

        code = "{self_param} + {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

    def __sub__(self, value):
        self._validate_type("__sub__")

        code = "{self_param} - {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

    def __div__(self, value):
        self._validate_type("__div__")

        code = "{self_param} / {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

    def __contain__(self, value):
        self._validate_type("__contain__")

        code = "{self_param} in {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

    def __equals__(self, value):
        self._validate_type("__equals__")

        code = "{self_param} == {value}"
        value_param = GraphParameter()
        value_param.set_value(value)
        new_param = GraphParameter()
        new_param.set_code_block(init_code=None, return_statement=code, self_param=self.parameter, value=value_param)
        return new_param

def unravel_list(lst):
    param = Value()
    param.set_value([])
    for val in lst:
        if type(val) in (list, tuple, set):
            val = unravel_list(val)
        elif type(val) is dict:
            val = unravel_dict(val)
        param.append(val)
    return param

def unravel_dict(dct):
    param = Value()
    param.set_value({})
    for key, val in dct.items():
        if type(val) in (list, tuple, set):
            val = unravel_list(val)
        elif type(val) is dict:
            val = unravel_dict(val)
        param[key] = val
    return param

def unravel(obj):
    if type(obj) in (list, tuple, set):
        return unravel_list(obj)
    elif type(obj) is dict:
        return unravel_dict(obj)
    else:
        raise TypeError("Wrong type: '%s', except list or dict" % type(obj))



from . import utils as ut
from . import custom_types
from . import graph_parameters as gp
from .the_exceptions import *

from .import *

from icecream import ic

class ValuePlaceholder(gp.Value):
    """A placeholder for a value return by a task"""
    def _init(self, run_job_id, name, worker_elf, *args, **kwargs):
        super(ValuePlaceholder, self)._init(*args, **kwargs)
        self.run_job_id = run_job_id
        self.name = name
        self.result_id = self.get_result_id(self.run_job_id, self.name)
        self.worker_elf = worker_elf

    def set_origin(self, result_id):
        """set the result id represented by the placeholder"""
        self.parameter.set_origin(self.result_id, self.worker_elf.mycelium.get_result)

    @classmethod
    def get_result_id(cls, job_id, name):
        return ut.legalize_key(job_id + name)

class TaskReturnPlaceHolder:
    """A place for the return of a task. Works as a dict, where each value is a key"""
    def __init__(self, worker_elf, task_function, run_job_id):
        self.task_function = task_function
        self.worker_elf = worker_elf
        self.run_job_id = run_job_id
        self.parameters = {}
        self.is_none = False
        
    def make_placeholder(self):
        """make and validate the placeholder"""
        from inspect import signature, _empty

        sig = signature(self.task_function)
        ret_annotation = sig.return_annotation

        self.is_none = False
        if (ret_annotation is _empty) :
            raise EmptyAnnotationError("Task return annotation cannot be empty, expecting None, dict, tuple or list of parameter keys. Got: '%s'" % ret_annotation)
        
        if (ret_annotation is None) :
            self.is_none = True
        elif (type(ret_annotation) not in [list, tuple, dict]) :
            raise EmptyAnnotationError("Task return annotation cannot be empty, expecting None, dict, tuple or list of parameter keys. Got: '%s'" % ret_annotation)

        if not self.is_none:
            if type(ret_annotation) is dict:
                for name, value in ret_annotation.items():
                    value_place = ValuePlaceholder(self.run_job_id, name, worker_elf=self.worker_elf, as_type=value)
                    value_place.set_origin(value_place.result_id)
                    self.parameters[name] = value_place
            else:
                for name in ret_annotation:
                    value_place = ValuePlaceholder(self.run_job_id, name, worker_elf=self.worker_elf)
                    value_place.set_origin(value_place.result_id)
                    self.parameters[name] = value_place

    def get_result_id(self, name):
        """return the result_id for value in the placeholder"""
        if name not in self.parameters:
            raise PlaceHolerKeyError("Placeholder has no parameter: '%s'" % name)

        return self.parameters[name].result_id
       
    def __getitem__(self, key):
        if self.is_none:
            return None
        return self.parameters[key]

    def __str__(self):
        if self.is_none:
            return "*-%s: None-*"% self.__class__.__name__

        return "*-%s: %s-*" %( self.__class__.__name__, str(self.parameters) )

    def __repr__(self):
        return str(self)

class TaskParameters:
    """Parameters of a task"""
    def __init__(self, fct, run_job_id, worker_elf):
        from inspect import signature

        self.run_job_id = run_job_id
        self.worker_elf = worker_elf
        self.signature = signature(fct)
        self.final_args = {}

    def set_parameters(self, *args, **kwargs):
        from inspect import _empty

        ret_annotation = self.signature.return_annotation
        parameters = self.signature.parameters
        self.final_args = {}
        
        for pid, param in enumerate(parameters):
            if pid < len(args):
                self.final_args[param] = args[pid]
            elif param in kwargs:
                if param in self.final_args:
                    raise RedundantParameterError("Got 2 values for parameter '{param}', ( {val1} & {val2})".format(param, self.final_args[param], kwargs[param]))
                self.final_args[param] = kwargs[param]
            else:
                if parameters[param].default is _empty:
                    raise RedundantParameterError("Mandatory parameter '{param}' missing)".format(param=param))
                self.final_args[param] = parameters[param].default
                
        return self.final_args
    
    def find_placeholders_in_key_value_iterrator(self, iterator):
        """find placeholders in dicts lists etc..."""
        placeholders = {}
        for key, value in iterator:
            if isinstance(value, ValuePlaceholder):
                placeholders[str(key)] = value
        return placeholders

    def validate(self):
        """returns placeholders"""
        from inspect import _empty

        accepted_types = [dict, list, tuple, int, float, bool]

        parameters = self.signature.parameters
        for param, value in self.final_args.items():
            if not isinstance(value, ValuePlaceholder):
                param_clas = None
                if not parameters[param].annotation is _empty :
                    param_clas = parameters[param].annotation
                elif not parameters[param].default is _empty :
                    param_clas = type(parameters[param].default)

                if not (param_clas is type(None)) and param_clas and not isinstance(value, param_clas):
                    raise ParameterTypeError("Param '{param}' has the wrong type {type} expected {exp})".format(param=param, type=type(value), exp=param_clas))

                if param_clas and (not param_clas in accepted_types ) :
                    raise ParameterTypeError("Param '{param}' has the wrong type {type} expected one of the following: {exp})".format(param=param, type=type(value), exp=accepted_types))

    # def get_storable_type(self, obj):
    #     """
    #     get a type storable by the mycellum for obj
    #         [list, tuple] -> array
    #         dict -> obj
    #         [str, float, int] -> primitive
    #         None -> null
    #     """
    #     if type(obj) in [list, tuple]:
    #         return "array"
    #     elif type(obj) is dict:
    #         return "object"
    #     elif type(obj) in [str, float, int]:
    #         return "primitive"
    #     elif type(obj) in type(None):
    #         return "null"
    #     else:
    #         raise Exception("Unknown type; %s must be in %s" % (type(obj), [list, tuple, dict, str, float, int]))

    def get_job_dependencies(self):
        """return the dependencies of a job (jobs that must run before) based on the parameters received"""
        def _rec(deps, iterator):
            for arg in iterator:
                if isinstance(arg, ValuePlaceholder):
                    deps.append(arg.run_job_id)
                elif type(arg) in [list, tuple, set]:
                    _rec(deps, arg)
                elif type(arg) is dict:
                    _rec(deps, arg.values())
        dependencies = []
        _rec(dependencies, self.final_args.values())
        return dependencies

    def get_parameter_dict(self):
        """return a dictionary of parameters"""
        params = {}
        for name, arg in self.final_args.items():
            if type(arg) in [dict, list, tuple, set]:
                param = gp.unravel(arg)
                val = ValuePlaceholder(self.run_job_id, name, worker_elf=self.worker_elf)
                val.parameter = param
            elif not isinstance(arg, ValuePlaceholder):
                val = ValuePlaceholder(self.run_job_id, name, worker_elf=self.worker_elf)
                val.set_value(arg)
            else:
                val = arg
            params[name] = val.traverse(to_dict=True)
        return params

    @classmethod
    def develop(cls, mycelium, dct_params:dict):
        """compute the value of parameters"""
        ret = {}
        for key, trav in dct_params.items():
            ret[key] = gp.GraphParameter.build_from_traversal(trav, pull_origin_function=mycelium.get_result)
            ret[key] = ret[key].make()
        return ret

class Job:
    """a Job for an elf"""
    def __init__(
        self,
        task,
        run_job_id,
        worker_elf,
        parameters,
        start_date,
        completion_date,
        status,
        mycelium,
        return_placeholder,
        dependencies
    ):
        self.task=task
        self.run_job_id=run_job_id
        self.worker_elf=worker_elf
        self.parameters=parameters
        self.start_date=start_date
        self.completion_date=completion_date
        self.status=status
        self.mycelium = mycelium
        self.return_placeholder=return_placeholder
        self.dependencies = dependencies

    def commit(self):
        """save the job to the mycelium"""
        self.mycelium.push_job(self)

class Task:
    """Represent a task for an elf"""
    def __init__(self, machine_elf, function, name):
        self.machine_elf = machine_elf
        self.function = function
        self.name = name
        # self.parameters = None
        self.return_placeholder = None

        self.source_code = None
        self.revision = None
        self.documentation = None
        self.signature = None
        
        self.inspect_function()

    def inspect_function(self):
        """inspect the function of a task"""
        import hashlib

        self.source_code = ut.inpsect_none_if_exception_or_empty(self.function, "getsource")
        self.revision = ut.legalize_key(self.name + ut.get_hash_key(self.source_code))
        self.documentation = ut.inpsect_none_if_exception_or_empty(self.function, "cleandoc")
        self.signature = ut.inpsect_none_if_exception_or_empty(self.function, "signature")

    def wrap(self, *args, **kwargs):
        """wrap the function inside a new function that creates a Job"""
        args = args
        kwargs = kwargs
        run_job_id = ut.legalize_key( self.name + ut.getuid() )
        # run_job_id = ut.getuid()

        def _wrapped():
            parameters = TaskParameters(self.function, run_job_id=run_job_id, worker_elf=self.machine_elf)
            parameters.set_parameters(*args, **kwargs)
            placeholders = parameters.validate()
            
            return_placeholder = TaskReturnPlaceHolder(worker_elf=self.machine_elf, task_function=self.function, run_job_id=run_job_id)
            return_placeholder.make_placeholder()

            dependencies = parameters.get_job_dependencies()
            # if len(placeholders) > 0:
                # dependencies = [phold.run_job_id for phold in placeholders]

            job = Job(
                task = self,
                run_job_id = run_job_id,
                worker_elf = self.machine_elf,
                parameters = parameters,
                start_date = None,
                completion_date = None,
                status = custom_types.STATUS["PENDING"],
                mycelium = self.machine_elf.mycelium,
                return_placeholder=return_placeholder,
                dependencies = dependencies
            )
            
            job_id = job.commit()

            return return_placeholder

        return _wrapped

    def run(self, job_id, *args, **kwargs):
        """run the task"""
        fct_ret = self.function(*args, **kwargs)
        
        if fct_ret is None:
            return None

        ret = {}
        for name, value_ret in fct_ret.items():
            ret[name] = {
                "result_id": ValuePlaceholder.get_result_id(job_id, name),
                "value": value_ret
            }

        return ret

    def __call__(self, *args, **kwargs):
        return self.wrap( *args, **kwargs )()

    def __str__(self):
        return "*-Task '%s.%s': %s -*" % (self.machine_elf.uid, self.name, self.signature)

    def __repr__(self):
        return str(self)

class MachineElf:
    """An elf that runs tasks"""

    def __init__(self, uid, mycelium):
        self.uid = ut.legalize_key(uid)
        self.mycelium = mycelium
        self.tasks = {}

        self.source_code = None
        self.revision = None
        self.documentation= None
        
        self.inspect_self()
        self.find_tasks()

    def inspect_self(self): 
        """inspect the elf to get the source code etc..."""
        self.source_code = ut.inpsect_none_if_exception_or_empty(self.__class__, "getsource")
        self.revision = ut.get_hash_key(self.source_code+self.uid, prefix=self.__class__.__name__ )
        self.documentation = ut.inpsect_none_if_exception_or_empty(self.__class__, "cleandoc")

    def find_tasks(self):
        """find all function whose name starts by `task_` """
        import inspect

        for name, member in inspect.getmembers(self):
            if name.startswith("task_") and inspect.ismethod(member):
                task = Task(self, member, name)
                self.tasks[name] = task
                setattr(self, name, task)

    def register(self, store_source=False):
        """register the elf to the mycelium"""
        self.mycelium.register_machine_elf(self, store_source)

    def get_jobs(self):
        """retuen the list of jobs for an elf"""
        return self.mycelium.get_jobs(self.uid)

    def is_job_ready(self, job_id):
        return self.mycelium.is_job_ready(job_id)

    def start_jobs(self, store_failures=True, raise_exceptions=True):
        """start a job for an elf"""
        jobs = self.mycelium.get_jobs(self.uid)
        for job in jobs:
            if self.mycelium.is_job_ready(job["id"]) :
                params = self.mycelium.get_job_parameters(job["id"])
                try:
                    params = TaskParameters.develop(self.mycelium, params)
                except ResultNotFound:
                    print(">>Unable to retrieve result. Will try again later.")
                else:
                    self.run_task(job["id"], job["task"]["name"], params, store_failures=store_failures, raise_exceptions=raise_exceptions)

    def run_task(self, job_id:str, task_name:str, parameters:dict, store_failures:bool, raise_exceptions:bool):
        """run a task for an elf"""
        import sys
        try:
            task = getattr(self, task_name)
            self.mycelium.start_job(job_id)
            ret = task.run(job_id, **parameters)
            self.mycelium.store_results(job_id, ret)
            self.mycelium.complete_job(job_id)
        except Exception as exp:
            if store_failures:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                self.mycelium.register_job_failure(exc_type, exc_value, exc_traceback, job_id)
            
            if raise_exceptions:
                raise exp

        return ret



import os
from setuptools import setup, find_packages

def get_all_filepaths_including_subdirs(path):
    paths = []
    for path, subdirs, files in os.walk(path):
        for name in files:
          paths.append(os.path.join(path, name))
    return paths
    
def get_package_version(here):
    import re
    filename = os.path.join(here, "blumycelium", "_version.py")
    with open(filename) as f:
        try:
            line = f.readlines()[0]
        except:
            raise RuntimeError("Impossible to read version name: %s" % filename)
    
    reg = r"^__version__\s*=\s*['\"]([^'\"]+)['\"]"
    res = re.search(reg, line, re.M)
    if res:
        version = res.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (filename,))

    return version.strip()

def read_file_content(here, name):
    with open(os.path.join(here, name)) as f:
        data = f.read()
    return data

here = os.path.abspath(os.path.dirname(__file__))

README = read_file_content(here, 'README.md')
CHANGES = read_file_content(here, 'CHANGES.md')
VERSION = get_package_version(here)

package_data = []
# package_data += get_all_filepaths_including_subdirs(here + '/blumycelium/locale')
# package_data += get_all_filepaths_including_subdirs(here + '/blumycelium/configuration_json')

requires = [
    "pyArango>=1.3.5",
    "icecream",
    "rich"
]

setup(name='blumycelium',
      version=VERSION,
      description='blumycelium',
      long_description=README,
      classifiers=[
        "Programming Language :: Python",
     ],
      author='Tariq Daouda',
      url='',
      keywords='arangodb multithread bluwr multiprocessing async',
      packages=find_packages(),
      include_package_data=True,
      package_data={'blumycelium': package_data},
      zip_safe=False,
      install_requires=requires,
      tests_require=requires,
      test_suite="blumycelium",
      entry_points="""\
      [paste.app_factory]
      main = blumycelium:main
      """,
      )




class EmptyParameter(Exception):
    """Represents an empty parameter"""
    pass

class EmptyAnnotationError(Exception):
    """"""
    pass

class PlaceHolerKeyError(Exception):
    """"""
    pass

class RedundantParameterError(Exception):
    """"""
    pass

class ParameterTypeError(Exception):
    """"""
    pass

class DatabaseNotFoundError(Exception):
    """"""
    pass


class ResultTypeError(Exception):
    """"""
    pass


class ValueDerivationError(Exception):
    """"""
    pass


class MultipleDerivationError(Exception):
    """"""
    pass


class DependencyTypeError(Exception):
    """"""
    pass


class ResultNotFound(Exception):
    """"""
    pass

